#include <spa/spa.h>

long SPA_Function1D::defaultMaxDepth = 15;
long SPA_Function1D::defaultMinDepth = 3;

SPA_Function1D::SPA_Function1D()
{
	rule = SPA_TRAPEZOID;
	accuracy = 1e-2;
	maxDepth = defaultMaxDepth;
	minDepth = defaultMinDepth;
	startDial = 1.0;
	tune = 0.1;
	maxtrap = 0.0;
}

SPA_Function1D::~SPA_Function1D()
{
}

void SPA_Function1D::Set(char r)
{
	rule = r;
}

void SPA_Function1D::Set(char r, double a)
{
	rule = r;
	accuracy = a;
}

void SPA_Function1D::Set(char r, double a, long max, long min)
{
	rule = r;
	accuracy = a;
	maxDepth = max;
	minDepth = min;
}

void SPA_Function1D::SetDefaultDepths(long max, long min)
{
	defaultMaxDepth = max;
	defaultMinDepth = min;
}

double SPA_Function1D::Dialin(double start, double end)
{
	double dial = startDial;
	double r1,r2;

	r1 = DialFunction(start, end, dial, 0.0);
	while(1)
	{
		dial *= tune;
		r2 = DialFunction(start, end, dial, 0.0);
		if(fabs(r2-r1) < accuracy)
		{
			break;
		}
		r1 = r2;
	}
	return r2;
}


double SPA_Function1D::DialFunction(double start, double end, double dial, double a0)
{
	maxtrap = 0.0;
	maxyval = 0.0;
	switch(rule)
	{
		case SPA_MIDYDIFF:
			return MidpointYdiff(start, end, dial, a0);
		default:
			wbaLog("SPA ERROR", "invalid rule type %d in DialFunction", rule);
			break;
	}
	return 0.0;
}

double SPA_Function1D::MidpointYdiff(double start, double end, double dial, double a0)
{
	double midpoint;
	double gap;
	double err, a1, a2, a3, x2, x3;

	midpoint = start + (end-start)/2.0;
	gap = midpoint - start;
	if(0.0 == a0)
	{
		a0 = 2.0*gap*Function(midpoint);
	}
	a1 = a0;
	x2 = start+gap/2.0;
	x3 = midpoint+gap/2.0;
	a2 = gap*Function(x2);
	a3 = gap*Function(x3);
	if(Function(x2) > maxyval){maxyval = Function(x2);}
	if(Function(x3) > maxyval){maxyval = Function(x3);}

	err = dial;

	if(fabs((x3-x2)/x2) > dial && !(fabs(x2)<1e-100 && fabs(x3)<1e-100))
	{
		//err = err/2.0;
		a2 = MidpointYdiff(start, midpoint, err, a2);
		a3 = MidpointYdiff(midpoint, end, err, a3);
	}
	else
	{
		if(a2 > maxtrap){maxtrap = a2;}
		if(a3 > maxtrap){maxtrap = a3;}
	}
	return a2+a3;
}

double SPA_Function1D::Rule(double a, double b)
{
	switch(rule)
	{
		case SPA_TRAPEZOID:
		case SPA_MIDPOINT:
		case SPA_MIDINF:
		case SPA_MIDEXP:
		case SPA_MIDUNIT:
			return Recipe(a, b);
		case SPA_MIDYDIFF:
			return Dialin(a, b);
		default:
			wbaLog("SPA ERROR", "unknown rule selected");
			break;
	}
	return 0.0;
}

double SPA_Function1D::Recipe(double a, double b)
{
	long j;
	double s, olds;
	
	olds = -1.0e30;
	//for(j=1;j<=maxDepth;j++)
	for(j=1;j>0;j++)
	{
		if(j > maxDepth)
		{
			wbaLog("SPA WARNING", "SPA_Function1D::Rule: depth of %d exceeded limit %d", j, maxDepth);
		}
		s=Choose(a,b,j);
		if(file.IsOpen())
		{
			fprintf(stderr,"*** %d %g\n", j, s);
			file.Printf("%d %g %g\n", j, s, b-a);
		}
		if(j>minDepth)
		{
			if(fabs(s-olds) < accuracy*fabs(olds) || (s == 0.0 && olds == 0.0))
			{
				return s;
			}
		}
		olds = s;
	}
	//wbaLog("SPA ERROR", "too many steps in SPA_Function1D::Rule");
	return 0.0;
}

double SPA_Function1D::Choose(double a, double b, long n)
{
	switch(rule)
	{
		case SPA_TRAPEZOID:
			return Trapezoid(a, b, n);
		case SPA_MIDPOINT:
			return Midpoint(a, b, n);
		case SPA_MIDINF:
			return MidInf(a, b, n);
		case SPA_MIDEXP:
			return MidExp(a, b, n);
		case SPA_MIDUNIT:
			return MidUnit(a, b, n);
		default:
			wbaLog("SPA ERROR", "invalid rule type %d in Choose", rule);
			break;
	}
	return 0.0;
}

double SPA_Function1D::Trapezoid(double a, double b, long n)
{
	double x,tnm,sum,del;
	long it,j;

	if(1 == n)
	{
		return (recipeS=0.5*(b-a)*(Function(a)+Function(b)));
	}
	else
	{
		for(it=1,j=1;j<n-1;j++) it <<= 1;
		tnm=it;
		del=(b-a)/tnm;
		x=a+0.5*del;
		for(sum=0.0,j=1;j<=it;j++,x+=del){sum += Function(x);}
		recipeS=0.5*(recipeS+(b-a)*sum/tnm);
		return recipeS;
	}
}
double SPA_Function1D::Midpoint(double a, double b, long n)
{
	double x, tnm, sum, del, ddel;
	long it, j;

	if(1 == n)
	{
		return (recipeS=(b-a)*Function(0.5*(a+b)));
	}
	else
	{
		for(it=1,j=1;j<n-1;j++) { it *= 3; }
		tnm = it;
		del = (b-a)/(3.0*tnm);
		ddel = del+del;
		x = a+0.5*del;
		sum=0.0;
		for(j=1;j<=it;j++)
		{
			sum += Function(x);
			x+=ddel;
			sum += Function(x);
			x+=del;
		}
		recipeS=(recipeS+(b-a)*sum/tnm)/3.0;
		return recipeS;
	}
}

double SPA_Function1D::FunctionInf(double x)
{
	return Function(1.0/(x))/((x)*(x));
}

double SPA_Function1D::MidInf(double aa, double bb, long n)
{
	double x, tnm, sum, del, ddel, a, b;
	long it, j;

	b = 1.0/(aa);
	a = 1.0/(bb);
	if(1 == n)
	{
		return (recipeS=(b-a)*FunctionInf(0.5*(a+b)));
	}
	else
	{
		for(it=1,j=1;j<n-1;j++) { it *= 3; }
		tnm = it;
		del = (b-a)/(3.0*tnm);
		ddel = del+del;
		x = a+0.5*del;
		sum=0.0;
		for(j=1;j<=it;j++)
		{
			sum += FunctionInf(x);
			x+=ddel;
			sum += FunctionInf(x);
			x+=del;
		}
		recipeS=(recipeS+(b-a)*sum/tnm)/3.0;
		return recipeS;
	}
}

double SPA_Function1D::FunctionExp(double x)
{
	return Function(-log(x))/(x);
}

double SPA_Function1D::MidExp(double aa, double bb, long n)
{
	double x, tnm, sum, del, ddel, a, b;
	long it, j;

	b = exp(-aa);
	a = 0.0;
	if(1 == n)
	{
		return (recipeS=(b-a)*FunctionExp(0.5*(a+b)));
	}
	else
	{
		for(it=1,j=1;j<n-1;j++) { it *= 3; }
		tnm = it;
		del = (b-a)/(3.0*tnm);
		ddel = del+del;
		x = a+0.5*del;
		sum=0.0;
		for(j=1;j<=it;j++)
		{
			sum += FunctionExp(x);
			x+=ddel;
			sum += FunctionExp(x);
			x+=del;
		}
		recipeS=(recipeS+(b-a)*sum/tnm)/3.0;
		return recipeS;
	}
}

double SPA_Function1D::FunctionUnit(double x, double a, double lambda)
{
	double oneminusx;
	oneminusx = 1.0-x;
	return 2.0*(lambda+a)*Function( (2.0*a + lambda*(x+1.0))/(oneminusx) ) / ((oneminusx)*(oneminusx));
}

double SPA_Function1D::MidUnit(double aa, double bb, long n)
{
	double x, tnm, sum, del, ddel, a, b;
	double lambda;
	long it, j;

	b = 1.0;
	a = -1.0;
	lambda = -aa + 0.1;
	if(1 == n)
	{
		return (recipeS=(b-a)*FunctionUnit(0.5*(a+b),aa,lambda));
	}
	else
	{
		for(it=1,j=1;j<n-1;j++) { it *= 3; }
		tnm = it;
		del = (b-a)/(3.0*tnm);
		ddel = del+del;
		x = a+0.5*del;
		sum=0.0;
		for(j=1;j<=it;j++)
		{
			sum += FunctionUnit(x,aa,lambda);
			x+=ddel;
			sum += FunctionUnit(x,aa,lambda);
			x+=del;
		}
		recipeS=(recipeS+(b-a)*sum/tnm)/3.0;
		return recipeS;
	}
}


